> module Merge where

> import Prelude hiding (pi,rem)
> import Abstract
> import List
> import Printing
> import Parsing (isWhite)

> 
> data Context                   = Context { pb     :: !Bool
>                                          , pem    :: !Bool
>                                          , pi     :: !Bool
>                                          , ps     :: !Bool
>                                          , ptt    :: !Bool
>                                          , pu     :: !Int
>                                          , psize  :: !Int
>                                          , pcolor :: !Color
>                                          }
>                                deriving Eq



> merge                        :: Context -> SMLNG -> SMLNG -> [SMLNG]
> merge _ (Text l) (Text r)      =  [Text (l ++ r)]
> merge _ l@(Text _) r@(Tag _ _) =  [r,l]
> merge c l@(Tag _ _) r@(Text _) =  reverse $ append c l r
> merge c l@(Tag tl bl) r@(Tag tr br)
>   | lti == ltl && lti == ltr =  [Tag tl (mergeL (modifyContext c tl)
>                                                 (reverse bl) br)]
>   | lti == ltl && (not r_haspl || l_haspl)
>                = [Tag tl (mergeL (modifyContext c tl) 
>                                  (reverse bl) [Tag (tr \\ t_inter) br])]
>   | lti == ltr && (not l_haspl || r_haspl)
>                = [Tag tr (mergeL (modifyContext c tr)
>                                  [Tag (tl \\ t_inter) bl] br)]
>   | lti > 0 && (not l_haspl && not r_haspl || l_haspl && r_haspl)  
>                = [Tag t_inter (append (modifyContext c t_inter) 
>                                       (Tag (tl \\ t_inter) bl)
>                                       (Tag (tr \\ t_inter) br))
>                  ]
>   | otherwise  = reverse $ append c l r
>  where 
>       {- we know, both tag sets have been preoptimised; if there is
>          a PL tag, then it is first -}
>       t_inter = tl `intersect` tr
>       lti     = length t_inter
>       ltl     = length tl
>       ltr     = length tr
>       l_haspl = not (null tl) && head tl == TPL
>       r_haspl = not (null tr) && head tr == TPL

> mergeL                       :: Context -> [SMLNG] -> [SMLNG] -> [SMLNG]
> mergeL c [] [x]                =  [x]
> mergeL c [] (x1:x2:xs)         =  mergeL c (merge c x1 x2) xs
> mergeL c fs []                 =  reverse fs
> mergeL c (f1:fr) (x1:xr)       =  mergeL c (merge c f1 x1 ++ fr) xr

> append                       :: Context -> SMLNG -> SMLNG -> [SMLNG]
> append c l@(Tag tl bl) r      
>   | tld == 0   = [Tag tl (mergeL (modifyContext c tl) (reverse bl) [r])]
>   | tld < length tl && not d_haspl
>                = [Tag t_new [Tag t_diff bl,r]]
>   | otherwise  = [l,r]
>  where nar = notAff_SMLNG c r
>        t_diff = tl `rem` nar
>        tld = length t_diff
>        d_haspl = TPL `elem` t_diff
>        t_new = tl \\ t_diff

> fullnotAff                   =  [TB,TEM,TI,TPL,TS,TTT,TU,Tsize (-1),Tcolor Cdef]

> notAff_String                :: Context -> String -> [Tag]
> notAff_String _ []           =  fullnotAff
> notAff_String c (x:xs)       =  (if   isWhite x
>                                  then (if pu c == 0 
>                                        then ((Tcolor Cdef):) else id
>                                       ) [TPL,TS,TB,TEM,TI]
>                                  else if ps c then [TEM] else []
>                                 ) `intersect` notAff_String c xs

> notAff_SMLNG                 :: Context -> SMLNG -> [Tag]
> notAff_SMLNG c (Tag t b)     =  foldr1 union 
>                                   ( notAff_SMLNGs (modifyContext c t) b 
>                                   : map tagNotAff t
>                                   )
> notAff_SMLNG c (Text t)      =  notAff_String c t

> notAff_SMLNGs                :: Context -> SMLNGs -> [Tag]
> notAff_SMLNGs c xs           =  foldr (\s ss -> notAff_SMLNG c s `intersect`
>                                                 ss) fullnotAff xs
> 
> rem :: [Tag] -> [Tag] -> [Tag]
> rem                   = foldl (flip $ deleteBy semieq)
> 
> semieq :: Tag -> Tag -> Bool
> semieq (Tcolor _) (Tcolor _) = True
> semieq (Tsize _) (Tsize _) = True
> semieq x y = x == y
> 					      
> modifyContext                :: Context -> [Tag] -> Context
> modifyContext                =  foldl mc
>   where mc c TB              =  c { pb = True }
>         mc c TEM             =  c { pem = if ps c then pem c else not (pem c) }
>         mc c TI              =  c { pi = True }
>         mc c TPL             =  c { pu = 0, pb = False, pem = False
>                                   , pi = False, ps = False, ptt = False }
>         mc c TS              =  c { ps = True }
>         mc c TTT             =  c { ptt = True }
>         mc c TU              =  c { pu = if (pu c<3) then pu c+1 else pu c } 
>         mc c (Tsize n)       =  c { psize = n }
>         mc c (Tcolor n)      =  c { pcolor = n }
> 
> 
> tagNotAff                    :: Tag -> [Tag]
> tagNotAff t                  =  case t of
>   TB                         -> [TB]
>   TI                         -> [TI]
>   TPL                        -> [TU,TB,TEM,TI,TS,TTT]
>   TS                         -> [TS,TEM]
>   TTT                        -> [TTT]
>   Tsize n                    -> [Tsize (-1)]
>   Tcolor n                   -> [Tcolor Cdef]
>   _                          -> []
> 

> plainContext                 :: Context -> Bool
> plainContext c               =  pu c == 0 && pb c == False && pem c == False
>                              && pi c == False && ps c == False 
>                              && ptt c == False

> impure                       :: Tag -> Bool
> impure (Tcolor _)            =  True
> impure (Tsize _)             =  True
> impure _                     =  False

> plintro                      :: Context -> SMLNG -> [SMLNG] -> [SMLNG] -> [SMLNG]
> plintro _ ch [] []             =  [ch]
> plintro c ch (s:ss) []         =  ch : plintro c s [] ss
> plintro c ch@(Text _) sk rs    =  ch : plintro c f [] r
>  where (f:r) = sk ++ rs
> plintro c ch sk@(s:ss) rs
>   | length sk > 2              =  ch : plintro c s [] (ss ++ rs)
> plintro c ch@(Tag _ _) [] (r:rs) = plintro c ch [r] rs
> plintro c ch@(Tag tl bl) sk (r@(Text _):rs) = plintro c ch (sk ++ [r]) rs
> plintro c ch@(Tag tl bl) sk (r@(Tag tr br):rs)
>   | length tip >= 2 && length tli == 0 = plintroL c
>                                        $ mergeL c 
>                                                 [Tag tl (bl++[Tag [TPL] sk])]
>                                                 (r:rs)
>   | length tip >= 2 = plintroL c
>                     $ mergeL c [Tag tlp [Tag tli bl,Tag [TPL] sk]] (r:rs)
>   | otherwise = plintro c ch (sk ++ [r]) rs
>  where t_inter = tl `intersect` tr
>        (tii,tip) = partition impure t_inter
>        (tli,tlp) = partition impure tl

> plintroL                     :: Context -> [SMLNG] -> [SMLNG]
> plintroL c (s:ss) | plainContext c = plintro c s [] ss
> plintroL c s                 =  s

