> module Parsing where

> import Parsec 
> import Char
> import Abstract

> open_tag                     :: Parser String
> open_tag                     =  do char '<'
>                                    name <- many1 alphaNum
>                                    char '>'
>                                    return name

> close_tag                    :: String -> Parser ()
> close_tag name               =  do string "</"
>                                    string name
>                                    char '>'
>                                    return ()

> transformtag                 :: String -> Parser Tag
> transformtag "B"             =  return TB
> transformtag "EM"            =  return TEM
> transformtag "I"             =  return TI
> transformtag "PL"            =  return TPL
> transformtag "S"             =  return TS
> transformtag "TT"            =  return TTT
> transformtag "U"             =  return TU
> transformtag "0"             =  return $ Tsize 0
> transformtag "1"             =  return $ Tsize 1
> transformtag "2"             =  return $ Tsize 2
> transformtag "3"             =  return $ Tsize 3
> transformtag "4"             =  return $ Tsize 4
> transformtag "5"             =  return $ Tsize 5
> transformtag "6"             =  return $ Tsize 6
> transformtag "7"             =  return $ Tsize 7
> transformtag "8"             =  return $ Tsize 8
> transformtag "9"             =  return $ Tsize 9
> transformtag "r"             =  return $ Tcolor Cr
> transformtag "g"             =  return $ Tcolor Cg
> transformtag "b"             =  return $ Tcolor Cb
> transformtag "c"             =  return $ Tcolor Cc
> transformtag "m"             =  return $ Tcolor Cm
> transformtag "y"             =  return $ Tcolor Cy
> transformtag "k"             =  return $ Tcolor Ck
> transformtag "w"             =  return $ Tcolor Cw
> transformtag _               =  fail "unknown tag"

> document                     :: Parser [SMLNG]
> document                     =  do blocks <- many (try block)
>                                    eof
>                                    return $ blocks

> block                        :: Parser SMLNG
> block                        =  tag <|> text

> tag                          :: Parser SMLNG
> tag                          =  do name <- open_tag
>                                    repr <- transformtag name
>                                    blocks <- many (try block)
>                                    close_tag name
>                                    return $ Tag [repr] blocks

> text                         :: Parser SMLNG
> text                         =  do txt <- many1 single
>                                    return $ Text txt

> single                       :: Parser Char
> single                       =  (satisfy isPrint' <?> "printable character")
>                              <|>(satisfy isWhite  <?> "whitespace")

> isWhite,isPrint'             :: Char -> Bool
> isWhite                      =  (`elem` " \r\n\t")
> isPrint' '<'                 =  False
> isPrint' '>'                 =  False
> isPrint' c                   =  isPrint c